from typing import Union

import nio
from nio.crypto import OlmDevice, TrustState
from slidge.command import Command, CommandAccess, Form, FormField, TableResult
from slidge.command.base import FormValues
from slidge.command.categories import CONTACTS, GROUPS
from slixmpp.exceptions import XMPPError

from .group import MUC
from .session import Session


class ListSpaces(Command):
    NAME = "🌌 Matrix spaces"
    CHAT_COMMAND = NODE = "spaces"
    CATEGORY = GROUPS
    HELP = "List the matrix spaces you're part of"
    ACCESS = CommandAccess.USER_LOGGED

    async def run(self, session: Session, _ifrom, *args: str) -> Form:  # type:ignore
        spaces = list[nio.MatrixRoom]()
        for room in session.matrix.rooms.values():
            if room.children:
                spaces.append(room)
        spaces = sorted(spaces, key=lambda r: r.name)
        return Form(
            title=self.NAME,
            instructions="Choose a space to list its children rooms. "
            "NB: as of now, you can also see rooms that you are a member of.",
            handler=self.finish,  # type:ignore
            handler_args=(spaces,),
            fields=[
                FormField(
                    "space",
                    label="Matrix space",
                    type="list-single",
                    options=[
                        {"label": room.display_name, "value": str(i)}
                        for i, room in enumerate(spaces)
                    ],
                )
            ],
        )

    @staticmethod
    async def finish(
        form_values: FormValues,
        session: Session,
        _ifrom,
        rooms: list[nio.MatrixRoom],
    ):
        space = rooms[int(form_values["space"])]  # type:ignore
        mucs = list[MUC]()
        for room_id in space.children:
            try:
                mucs.append(await session.bookmarks.by_legacy_id(room_id))
            except XMPPError:
                continue

        mucs = sorted(mucs, key=lambda muc: muc.name or "Unnamed room")
        return TableResult(
            fields=[FormField("name"), FormField("jid", type="jid-single")],
            description=f"Rooms of '{space.display_name}'",
            jids_are_mucs=True,
            items=[
                {"name": muc.name or "Unnamed room", "jid": str(muc.jid)}
                for muc in mucs
            ],
        )


class ManageTrust(Command):
    NAME = "🤝 Manage trust in devices"
    CATEGORY = CONTACTS
    CHAT_COMMAND = NODE = "verify"
    HELP = "Manage which OLM keys you trust or not."
    ACCESS = CommandAccess.USER_LOGGED

    HUMAN_STATES = {0: "unset", 1: "verified", 2: "blacklisted", 3: "ignored"}

    def __human_device(self, d: OlmDevice, state=True):
        r = f"{d.ed25519} of {d.user_id}"
        if state:
            return r + f" ({self.HUMAN_STATES[d.trust_state.value]})"
        return r

    async def run(
        self,
        session: Session,  # type:ignore
        _ifrom,
        *args: str,
    ) -> Union[Form, str]:
        devices = list[OlmDevice](session.matrix.olm.device_store)
        device_dict = {d.id: d for d in devices}

        # this part if for chat commands only
        if args:
            if args[0] == "all":
                return await self.step2(
                    {
                        "device": list(device_dict.keys()),  # type:ignore
                        "new_state": "verified",
                    },
                    session,
                    None,
                    device_dict,
                )

            else:
                return await self.step2(
                    {"device": args[0].upper(), "new_state": "verified"},
                    session,
                    None,
                    device_dict,
                )

        return Form(
            title=self.NAME,
            instructions="Choose the session(s) which trust state you want to change",
            handler=self.step2,  # type:ignore
            handler_args=(device_dict,),
            fields=[
                FormField(
                    "device",
                    label="Device(s)",
                    type="list-multi",
                    options=[
                        {"label": self.__human_device(d), "value": d.id}
                        for d in devices
                    ],
                ),
                FormField(
                    "new_state",
                    label="What new status do you want to give the selected devices?",
                    type="list-single",
                    options=[
                        {"label": v, "value": v} for v in self.HUMAN_STATES.values()
                    ],
                ),
            ],
        )

    async def step2(
        self,
        form_values: FormValues,
        session: Session,
        _ifrom,
        devices: dict[str, OlmDevice],
    ):
        new_state = form_values["new_state"]
        matrix = session.matrix
        result = ""
        for device_name in form_values["device"]:  # type:ignore
            device = devices[device_name]
            change = False
            if new_state == "unset":
                if device.trust_state == TrustState.verified:
                    change = matrix.unverify_device(device)
                elif device.trust_state == TrustState.ignored:
                    change = matrix.unignore_device(device)
                elif device.trust_state == TrustState.blacklisted:
                    change = matrix.unblacklist_device(device)
            elif new_state == "verified":
                change = session.matrix.verify_device(device)
            elif new_state == "blacklisted":
                change = session.matrix.blacklist_device(device)
            elif new_state == "ignored":
                change = session.matrix.ignore_device(device)

            if change:
                result += (
                    f"\nThe status of {self.__human_device(device, False)} "
                    f"is now {new_state}."
                )
            else:
                result += (
                    f"\nThe status of {self.__human_device(device, False)} "
                    f"has not changed."
                )

        return result or "Nothing was changed."



MAX_HISTORY_FETCH = 100
MAX_HISTORY_FETCH__DOC = (
    "Number of events to fetch to back-fill MUC history before slidge starts up."
)

MAX_PARTICIPANTS_FETCH = 10
MAX_PARTICIPANTS_FETCH__DOC = (
    "Number of participants to fetch when joining a group. "
    "Higher values will make joining slower, and participants will appear when "
    "they speak or if they spoke in the back-filled events. "
    "Participants with power levels > 50 (ie, admins) will be fetched."
)

NIO_VERBOSE = False
NIO_VERBOSE__DOC = (
    "Set this to True to respect the global log level from the matrix lib. "
    "It's REALLY verbose, so the default is to use WARNING."
)

PARSE_MESSAGE_STYLING = True
PARSE_MESSAGE_STYLING__DOC = """
Convert Message Styling (XEP-0393) message bodies to Matrix custom HTML.
Supported markup:
_underline_
*bold*
~strikethrough~
`code span`
```code block```
>quote
||spoiler
\\_escape style_
"""

REACTION_CACHE_SIZE = 10000
REACTION_CACHE_SIZE__DOC = "Emoji reaction cache size, in number of messages"



import typing

import nio
from slidge.group import LegacyBookmarks, LegacyMUC, LegacyParticipant, MucType
from slidge.util.types import HoleBound
from slixmpp.exceptions import XMPPError

from . import config
from .util import MatrixMixin, server_timestamp_to_datetime

if typing.TYPE_CHECKING:
    from .session import Session


class Participant(MatrixMixin, LegacyParticipant):
    session: "Session"
    muc: "MUC"

    def set_affiliation_from_power_level(self, power_level: int) -> None:
        if power_level == 100:
            self.affiliation = "owner"
            self.role = "moderator"
        elif power_level >= 50:
            self.affiliation = "admin"
            self.role = "moderator"
        else:
            self.affiliation = "member"
            self.role = "participant"


class Bookmarks(LegacyBookmarks):
    session: "Session"

    async def fill(self):
        self.log.debug("Filling rooms")
        for room in self.session.matrix.rooms:
            try:
                await self.by_legacy_id(room)
            except XMPPError as e:
                self.log.debug(
                    "%s is not a group chat or trouble getting it: %r", room, e
                )


class MUC(LegacyMUC[str, str, Participant, str]):
    session: "Session"
    type = MucType.CHANNEL_NON_ANONYMOUS

    async def get_room(self):
        try:
            return self.session.matrix.rooms[self.legacy_id]
        except KeyError:
            raise XMPPError("item-not-found", f"No room named {self.legacy_id}")

    async def get_message(self, msg_id: str) -> typing.Optional[nio.Event]:
        return await self.session.matrix.get_event(self.legacy_id, msg_id)

    async def update_info(self):
        room = await self.get_room()

        if new := room.replacement_room:
            localpart = await self.session.bookmarks.legacy_id_to_jid_username(new)
            raise XMPPError(
                "redirect", f"xmpp:{localpart}@{self.xmpp.boundjid.bare}?join"
            )
        self.log.debug("Children: %s", room.children)
        if room.children:
            raise XMPPError("bad-request", "This is not a real room but a 'space'")

        self.user_nick = room.user_name(self.session.matrix.user_id)

        self.log.debug("User nick: %s", self.user_nick)
        self.name = room.display_name
        self.log.debug("Avatar: %s", room.room_avatar_url)
        self.n_participants = room.member_count
        self.subject = room.topic

        if room.room_avatar_url:
            self.avatar = await self.session.matrix.mxc_to_http(room.room_avatar_url)

        elif room.member_count == 2:
            # if 1:1 set room avatar to other users's avatar
            for user_id, user in list(room.users.items()):
                if user_id == self.session.matrix.user_id:
                    continue
                if user.avatar_url is None:
                    break
                self.avatar = await self.session.matrix.mxc_to_http(user.avatar_url)
                break

    async def fill_participants(self):
        room = await self.get_room()
        if not room.members_synced:
            resp = await self.session.matrix.joined_members(self.legacy_id)
            if isinstance(resp, nio.JoinedMembersResponse):
                self.log.debug(
                    "Joined members response: %s participants", len(resp.members)
                )
            else:
                self.log.debug("Joined members error: %s", resp)

        power_levels = room.power_levels.users
        i = 0

        for user_id, user in list(room.users.items()):
            power_level = power_levels.get(user_id, 0)
            if power_level < 50 and i > config.MAX_PARTICIPANTS_FETCH:
                continue
            try:
                p = await self.get_participant_by_legacy_id(user.user_id)
            except XMPPError:
                continue
            p.set_affiliation_from_power_level(power_level)
            if power_level < 50:
                i += 1
            yield p

    async def backfill(
        self,
        after: HoleBound | None = None,
        before: HoleBound | None = None,
    ):
        after_date = None if after is None else after.timestamp
        before_date = None if before is None else before.timestamp
        async for event in self.session.matrix.fetch_history(
            self.legacy_id,
            config.MAX_HISTORY_FETCH,
        ):
            if not isinstance(event, nio.RoomMessage):
                self.log.debug("Not back-filling with %s", type(event))
            when = server_timestamp_to_datetime(event)
            if after_date is not None and when <= after_date:
                continue
            if before_date is not None and when >= before_date:
                continue
            try:
                participant = await self.session.matrix.get_participant(
                    await self.get_room(), event
                )
            except XMPPError as e:
                # maybe deleted profiles?
                self.log.debug(
                    "Something is wrong with participant %s, falling back to using matrix ID as nickname: %s",
                    event.sender,
                    e,
                )
                participant = await self.get_participant(event.sender)

            await participant.send_matrix_message(event, archive_only=True)

            # FIXME: this breaks everything, probably because of parallel calls
            #        to sync()
            # if isinstance(event, nio.UnknownEvent) and event.type == "m.reaction":
            #     await self.session.matrix.on_reaction(
            #         await self.get_room(), event, when=when, archive_only=True
            #     )
            #     continue



import asyncio
from datetime import datetime, timedelta
from typing import TYPE_CHECKING

import nio
from aiohttp import ClientConnectionError
from slidge import LegacyContact, LegacyRoster
from slixmpp.exceptions import XMPPError

if TYPE_CHECKING:
    from .session import Session


class Contact(LegacyContact[str]):
    """
    We don't implement direct messages but the what's parsed here will propagate
    to MUC participants.
    """

    session: "Session"

    async def update_info(self):
        # Because of slidge core's design, at this point we don't know which room
        # the contact is in, so we have to iterate over rooms, which is better than
        # issuing a full profile fetch
        for _, room in self.session.matrix.rooms.items():
            user = room.users.get(self.legacy_id)
            if user is not None:
                break
        else:
            raise XMPPError("item-not-found", "We don't know this user")

        self.name = user.display_name
        if mxc := user.avatar_url:
            self.avatar = await self.session.matrix.mxc_to_http(mxc)
        else:
            self.avatar = None

    async def fetch_vcard(self):
        try:
            resp = await self.session.matrix.get_profile(self.legacy_id)
        except (ClientConnectionError, TimeoutError, asyncio.TimeoutError) as e:
            raise XMPPError("remote-server-timeout", str(e))
        if not isinstance(resp, nio.ProfileGetResponse):
            raise XMPPError("internal-server-error", str(resp))

        if resp.other_info:
            self.set_vcard(note=str(resp.other_info))

    def update_presence(self, p: nio.PresenceEvent):
        kw = dict(status=p.status_msg)
        if last := p.last_active_ago is not None:
            kw["last_seen"] = datetime.now() - timedelta(seconds=last)
        if p.currently_active:
            self.online(**kw)
        else:
            self.away(**kw)


class Roster(LegacyRoster[str, Contact]):
    session: "Session"

    async def jid_username_to_legacy_id(self, jid_username: str):
        u = await super().jid_username_to_legacy_id(jid_username)
        if not u.startswith("@"):
            raise XMPPError(
                "bad-request",
                f"'{jid_username}' is not a valid matrix username. "
                "Matrix usernames starts with @",
            )
        return u



import nio

from matridge.util import strip_reply_fallback, get_body


def test_reply_to():
    reply = {
        "content": {
            "body": "> dddd\n\n * for update a list of users i modified the updategroupparticipants functions",
            "format": "org.matrix.custom.html",
            "formatted_body": '<mx-reply><blockquote><a href="https://matrix.to/#/!XXXX.net&amp;via=matrix.org&amp;via=aria-net.org">In reply to</a> <a href="https://matrix.to/#/@XX">@SDASDA:matrix.org</a><br>'
            "aasdasdasd"
            "</blockquote></mx-reply> * for update a list of users i modified the updategroupparticipants functions",
            "m.new_content": {
                "body": "for update a list of users i modified the updategroupparticipants functions",
                "format": "org.matrix.custom.html",
                "formatted_body": "for update a list of users i modified the updategroupparticipants functions",
                "msgtype": "m.text",
            },
            "m.relates_to": {
                "event_id": "asdasda",
                "rel_type": "m.replace",
            },
            "msgtype": "m.text",
        },
        "origin_server_ts": 1688997470785,
        "sender": "@asdasd:matrix.org",
        "type": "m.room.message",
        "unsigned": {"age": 75665575},
        "event_id": "asdasd",
        "room_id": "!whatsmeow-v9:maunium.net",
    }
    event = nio.Event.parse_event(reply)
    assert (
        get_body(event)
        == " * for update a list of users i modified the updategroupparticipants functions"
    )
    reply = {
        "content": {
            "body": "> <@nicocool:matrix.org> asdfasdf\n\nsadfsadf",
            "format": "org.matrix.custom.html",
            "formatted_body": '<mx-reply><blockquote><a href="https://matrix.to/#/!fFWFUtcegfROhIEMbt:matrix.org/$gdL7K5fh98eQTAl_EUdFkvH2xskV0UCO5i-vUeVY3Zk?via=matrix.org">In reply to</a> <a href="https://matrix.to/#/@nicocool:matrix.org">@nicocool:matrix.org</a><br>asdfasdf</blockquote></mx-reply>sadfsadf',
            "m.relates_to": {
                "m.in_reply_to": {
                    "event_id": "$gdL7K5fh98eQTAl_EUdFkvH2xskV0UCO5i-vUeVY3Zk"
                }
            },
            "msgtype": "m.text",
        },
        "origin_server_ts": 1689102822908,
        "sender": "@nicovraimentcool:matrix.org",
        "type": "m.room.message",
        "unsigned": {"age": 710},
        "event_id": "$RAOwtjYb487xbVc9zmVJN_rLVLeSKLoxmGyucp9BsL4",
    }
    event = nio.Event.parse_event(reply)
    assert get_body(event) == "sadfsadf"


def test_strip_reply_fallback():
    assert (
        strip_reply_fallback(  # language=html
            "<mx-reply>"
            '<blockquote><a href="https://matrix.to/XXX">In reply to</a> <a href="https://matrix.to/#/@nicovraimentcool:matrix.org">@nicovraimentcool:matrix.org</a><br>fdg</blockquote>'
            "</mx-reply>"
            "dfggg"
        )
        == "dfggg"
    )


def test_notice():
    notice = {
        "content": {
            "body": "**[mautrix/facebook]** sdfsafasdfsadf starred the repo",
            "com.beeper.linkpreviews": [],
            "format": "org.matrix.custom.html",
            "formatted_body": '<strong>[<a data-mautrix-exclude-plaintext href="https://github.com/mautrix/facebook">mautrix/facebook</a>]</strong> <a data-mautrix-exclude-plaintext href="https://github.com/sdfsafasdfsadf">sdf</a> starred the repo',
            "msgtype": "m.notice",
            "xyz.maubot.github.webhook": {
                "delivery_ids": ["asdfasdf-1d28-11ee-9be9-fec185c673cf"],
                "event_type": "star",
            },
        },
        "origin_server_ts": 1688776979579,
        "room_id": "!facebook-v10:maunium.net",
        "sender": "@github:maunium.net",
        "type": "m.room.message",
        "unsigned": {},
        "event_id": "sdfasdf",
        "user_id": "@github:maunium.net",
    }
    event = nio.Event.parse_event(notice)
    assert get_body(event) == "**[mautrix/facebook]** sdfsafasdfsadf starred the repo"
    notice = {
        "content": {
            "body": "**[tulir/whatsmeow]** tulir pushed [1 commit](https://github.com/tulir/whatsmeow/compare/6e8b189f1308...93091c7024da) to main:\n● `93091c70` Fix decrypting message secret reactions",
            "com.beeper.linkpreviews": [],
            "format": "org.matrix.custom.html",
            "formatted_body": '<strong>[<a data-mautrix-exclude-plaintext href="https://github.com/tulir/whatsmeow">tulir/whatsmeow</a>]</strong> <a data-mautrix-exclude-plaintext href="https://github.com/tulir">tulir</a>      pushed\n        <a href="https://github.com/tulir/whatsmeow/compare/6e8b189f1308...93091c7024da">1 commit</a>\n    to\n main:<ul>         <li>\n            <code><a data-mautrix-exclude-plaintext href="https://github.com/tulir/whatsmeow/commit/93091c7024dac73790fafe4036f6771352d40ca0">93091c70</a></code>\n            Fix decrypting message secret reactions\n        </li>\n </ul>',
            "msgtype": "m.notice",
            "xyz.maubot.github.webhook": {
                "delivery_ids": ["7e0044f0-1f0b-11ee-9cbd-cba6b984b843"],
                "event_type": "push",
            },
        },
        "origin_server_ts": 1688984472208,
        "room_id": "!whatsmeow-v9:maunium.net",
        "sender": "@github:maunium.net",
        "type": "m.room.message",
        "unsigned": {},
        "event_id": "$-ff",
        "user_id": "@github:maunium.net",
    }
    event = nio.Event.parse_event(notice)
    assert (
        get_body(event)
        == "**[tulir/whatsmeow]** tulir pushed [1 commit](https://github.com/tulir/whatsmeow/compare/6e8b189f1308...93091c7024da) to main:\n● `93091c70` Fix decrypting message secret reactions"
    )



# DO NOT EDIT
# Generated from .copier-answers.yml

import sys
from importlib.metadata import PackageNotFoundError, version

from slidge import entrypoint

# import everything for automatic subclasses discovery by slidge core
from . import command, config, contact, gateway, group, session

try:
    __version__ = version("matridge")
except PackageNotFoundError:
    # package is not installed
    __version__ = "dev"


def main():
    if len(sys.argv) == 2 and sys.argv[1] == "--version":
        print("matridge version", __version__)
        exit(0)
    entrypoint("matridge")


__all__ = (
    "__version__",
    "command",
    "config",
    "contact",
    "gateway",
    "group",
    "main",
    "session",
)
